SlopeCompensationExtended = {
	MOD_NAME = g_currentModName,
	SPEC_NAME = g_currentModName .. ".slopeCompensationExtended"
}
SlopeCompensationExtended.SPEC_TABLE_NAME = "spec_" .. SlopeCompensationExtended.SPEC_NAME
SlopeCompensationExtended.SLOPE_COLLISION_MASK = 223

function SlopeCompensationExtended.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Wheels, specializations)
end

function SlopeCompensationExtended.initSpecialization()
	g_configurationManager:addConfigurationType("slopeCompensationExtended", g_i18n:getText("shop_configuration"), "slopeCompensationExtended", nil, nil, nil, ConfigurationUtil.SELECTOR_MULTIOPTION)

	local schema = Vehicle.xmlSchema

	schema:setXMLSpecializationType("SlopeCompensationExtended")
	SlopeCompensationExtended.registerXMLPaths(schema, "vehicle.slopeCompensationExtended")
	SlopeCompensationExtended.registerXMLPaths(schema, "vehicle.slopeCompensationExtended.slopeCompensationExtendedConfigurations.slopeCompensationExtendedConfiguration(?)")
	ObjectChangeUtil.registerObjectChangeXMLPaths(schema, "vehicle.slopeCompensationExtended.slopeCompensationExtendedConfigurations.slopeCompensationExtendedConfiguration(?)")
	schema:addDelayedRegistrationFunc("AnimatedVehicle:part", function (cSchema, cKey)
		cSchema:register(XMLValueType.INT, cKey .. "#slopeCompensationNodeIndex", "Index in the XML of the slope compensation node")
		cSchema:register(XMLValueType.FLOAT, cKey .. "#startSlopeCompensationLevel", "Start slope compensation level")
		cSchema:register(XMLValueType.FLOAT, cKey .. "#endSlopeCompensationLevel", "End slope compensation level")
	end)
	schema:setXMLSpecializationType()

	local schemaSavegame = Vehicle.xmlSchemaSavegame

	schemaSavegame:register(XMLValueType.FLOAT, string.format("vehicles.vehicle(?).%s.compensationNode(?)#position", SlopeCompensationExtended.SPEC_NAME), "Current compensation node position")
end

function SlopeCompensationExtended.registerXMLPaths(schema, basePath)
	schema:register(XMLValueType.FLOAT, basePath .. "#threshold", "Update threshold for animation", 0.002)
	schema:register(XMLValueType.BOOL, basePath .. "#highUpdateFrequency", "Defines if the angle is updated every frame or every seconds frame", false)
	schema:register(XMLValueType.NODE_INDEX, basePath .. ".compensationNode(?)#wheelNode1", "Wheel node 1")
	schema:register(XMLValueType.NODE_INDEX, basePath .. ".compensationNode(?)#wheelNode2", "Wheel node 2")
	schema:register(XMLValueType.ANGLE, basePath .. ".compensationNode(?)#maxAngle", "Max. angle", 5)
	schema:register(XMLValueType.ANGLE, basePath .. ".compensationNode(?)#minAngle", "Min. angle", "Negative #maxAngle")
	schema:register(XMLValueType.FLOAT, basePath .. ".compensationNode(?)#speed", "Move speed", 1)
	schema:register(XMLValueType.FLOAT, basePath .. ".compensationNode(?)#inActiveHeight", "Height while the compensation node is not active", 0)
	schema:register(XMLValueType.FLOAT, basePath .. ".compensationNode(?)#initialHeight", "Height while the componensation node is active and the vehicle is fully leveled", 1)
	schema:register(XMLValueType.NODE_INDEX, basePath .. ".compensationNode(?).animationPart(?)#node", "Node that is adjusted")
	schema:register(XMLValueType.BOOL, basePath .. ".compensationNode(?).animationPart(?)#isLeft", "Is left or right side node")
	schema:register(XMLValueType.VECTOR_ROT, basePath .. ".compensationNode(?).animationPart(?)#rotMin", "Min. rotation")
	schema:register(XMLValueType.VECTOR_ROT, basePath .. ".compensationNode(?).animationPart(?)#rotMax", "Max. rotation")
	schema:register(XMLValueType.VECTOR_TRANS, basePath .. ".compensationNode(?).animationPart(?)#transMin", "Min. translation")
	schema:register(XMLValueType.VECTOR_TRANS, basePath .. ".compensationNode(?).animationPart(?)#transMax", "Max. translation")
end

function SlopeCompensationExtended.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "loadCompensationNodeExtFromXML", SlopeCompensationExtended.loadCompensationNodeExtFromXML)
	SpecializationUtil.registerFunction(vehicleType, "updateSlopeCompensationNodeExt", SlopeCompensationExtended.updateSlopeCompensationNodeExt)
	SpecializationUtil.registerFunction(vehicleType, "getCompensationGroundPositionExt", SlopeCompensationExtended.getCompensationGroundPositionExt)
end

function SlopeCompensationExtended.registerEventListeners(vehicleType)
	SpecializationUtil.registerEventListener(vehicleType, "onLoad", SlopeCompensationExtended)
	SpecializationUtil.registerEventListener(vehicleType, "onLoadFinished", SlopeCompensationExtended)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdate", SlopeCompensationExtended)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", SlopeCompensationExtended)
	SpecializationUtil.registerEventListener(vehicleType, "onRegisterAnimationValueTypes", SlopeCompensationExtended)
end

function SlopeCompensationExtended:onLoad(savegame)
	local spec = self[SlopeCompensationExtended.SPEC_TABLE_NAME]
	local configurationId = Utils.getNoNil(self.configurations.slopeCompensationExtended, 1)
	local configKey = string.format("vehicle.slopeCompensationExtended.slopeCompensationExtendedConfigurations.slopeCompensationExtendedConfiguration(%d)", configurationId - 1)

	ObjectChangeUtil.updateObjectChanges(self.xmlFile, "vehicle.slopeCompensationExtended.slopeCompensationExtendedConfigurations.slopeCompensationExtendedConfiguration", configurationId, self.components, self)

	if not self.xmlFile:hasProperty(configKey) then
		configKey = "vehicle.slopeCompensationExtended"
	end

	spec.lastRaycastDistance = 0
	spec.nodes = {}

	self.xmlFile:iterate(configKey .. ".compensationNode", function (index, key)
		local compensationNode = {}

		if self:loadCompensationNodeExtFromXML(compensationNode, self.xmlFile, key) then
			table.insert(spec.nodes, compensationNode)
		end
	end)

	if savegame ~= nil and not savegame.resetVehicles then
		for i, compensationNode in ipairs(spec.nodes) do
			local nodeKey = string.format("%s.%s.compensationNode(%d)", savegame.key, SlopeCompensationExtended.SPEC_NAME, i - 1)

			if savegame.xmlFile:hasProperty(nodeKey) then
				compensationNode.lastPosition = savegame.xmlFile:getValue(nodeKey .. "#position", compensationNode.lastPosition)

				self:updateSlopeCompensationNodeExt(compensationNode, compensationNode.lastPosition, 99999)
			end
		end
	end

	spec.threshold = self.xmlFile:getValue(configKey .. "#threshold", 0.002)

	if #spec.nodes == 0 then
		SpecializationUtil.removeEventListener(self, "onLoadFinished", SlopeCompensationExtended)
		SpecializationUtil.removeEventListener(self, "onUpdate", SlopeCompensationExtended)
		SpecializationUtil.removeEventListener(self, "onUpdateTick", SlopeCompensationExtended)
	elseif self.xmlFile:getValue(configKey .. "#highUpdateFrequency", false) then
		SpecializationUtil.removeEventListener(self, "onUpdateTick", SlopeCompensationExtended)
	else
		SpecializationUtil.removeEventListener(self, "onUpdate", SlopeCompensationExtended)
	end
end

function SlopeCompensationExtended:onLoadFinished(savegame)
	local spec = self[SlopeCompensationExtended.SPEC_TABLE_NAME]

	for i, compensationNode in ipairs(spec.nodes) do
		self:updateSlopeCompensationNodeExt(compensationNode, compensationNode.lastPosition, 99999)
	end
end

function SlopeCompensationExtended:saveToXMLFile(xmlFile, key, usedModNames)
	local spec = self[SlopeCompensationExtended.SPEC_TABLE_NAME]

	for i, compensationNode in ipairs(spec.nodes) do
		local nodeKey = string.format("%s.compensationNode(%d)", key, i - 1)

		xmlFile:setValue(nodeKey .. "#position", compensationNode.lastPosition)
	end
end

function SlopeCompensationExtended:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local spec = self[SlopeCompensationExtended.SPEC_TABLE_NAME]

	for _, compensationNode in ipairs(spec.nodes) do
		local x1, y1, z1, valid1 = self:getCompensationGroundPositionExt(compensationNode, 1)
		local x2, y2, z2, valid2 = self:getCompensationGroundPositionExt(compensationNode, 2)

		if valid1 and valid2 then
			local h = y1 - y2
			local l = MathUtil.vector2Length(x1 - x2, z1 - z2)
			local angle = math.tan(h / l)
			local position = MathUtil.clamp((angle - compensationNode.minAngle) / (compensationNode.maxAngle - compensationNode.minAngle), 0, 1)

			if spec.threshold < math.abs(compensationNode.lastPosition - position) then
				local dir = MathUtil.sign(position - compensationNode.lastPosition)
				local limit = dir > 0 and math.min or math.max
				compensationNode.lastPosition = limit(compensationNode.lastPosition + compensationNode.speed * dt * dir, position)

				self:updateSlopeCompensationNodeExt(compensationNode, compensationNode.lastPosition, dt)
			end
		end
	end
end

function SlopeCompensationExtended:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	SlopeCompensationExtended.onUpdate(self, dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
end

function SlopeCompensationExtended:onRegisterAnimationValueTypes()
	self:registerAnimationValueType("slopeCompensationLevel", "startSlopeCompensationLevel", "endSlopeCompensationLevel", false, AnimationValueFloat, function (value, xmlFile, xmlKey)
		value.slopeCompensationNodeIndex = xmlFile:getValue(xmlKey .. "#slopeCompensationNodeIndex")

		if value.slopeCompensationNodeIndex ~= nil then
			value:setWarningInformation("index: " .. tostring(value.slopeCompensationNodeIndex))
			value:addCompareParameters("slopeCompensationNodeIndex")

			return true
		end

		return false
	end, function (value)
		if value.slopeCompensationNode == nil then
			local spec = value.vehicle[SlopeCompensationExtended.SPEC_TABLE_NAME]
			local slopeCompensationNode = spec.nodes[value.slopeCompensationNodeIndex]

			if slopeCompensationNode == nil then
				Logging.xmlWarning(value.xmlFile, "Could not update slope compensation node level. No slope compensation node with index %d found!", value.slopeCompensationNodeIndex)

				value.startValue = nil

				return 0
			end

			value.slopeCompensationNode = slopeCompensationNode
		end

		return value.slopeCompensationNode.compensationLevel
	end, function (value, compensationLevel)
		if value.slopeCompensationNode ~= nil then
			value.slopeCompensationNode.compensationLevel = compensationLevel

			self:updateSlopeCompensationNodeExt(value.slopeCompensationNode, value.slopeCompensationNode.lastPosition, 99999)
		end
	end)
end

function SlopeCompensationExtended:loadCompensationNodeExtFromXML(compensationNode, xmlFile, key)
	compensationNode.raycastDistance = 0

	for i = 1, 2 do
		local wheelNode = xmlFile:getValue(key .. "#wheelNode" .. tostring(i), nil, self.components, self.i3dMappings)

		if wheelNode == nil then
			Logging.xmlWarning(xmlFile, "Missing wheelNode%d for compensation node '%s'", i, key)

			return false
		else
			local wheel = self:getWheelByWheelNode(wheelNode)

			if wheel ~= nil then
				compensationNode[string.format("wheel%dNode", i)] = wheel.driveNode
				compensationNode.raycastDistance = math.max(compensationNode.raycastDistance, wheel.radius + 1)
			else
				Logging.xmlWarning(xmlFile, "Unknown wheel node for compensation node '%s'", key)

				return false
			end
		end
	end

	compensationNode.maxAngle = xmlFile:getValue(key .. "#maxAngle", 5)
	compensationNode.minAngle = xmlFile:getValue(key .. "#minAngle", -math.deg(compensationNode.maxAngle))
	compensationNode.speed = xmlFile:getValue(key .. "#speed", 1) / 1000
	compensationNode.inActiveHeight = xmlFile:getValue(key .. "#inActiveHeight", 0)
	compensationNode.initialHeightOffset = 1 - MathUtil.clamp(xmlFile:getValue(key .. "#initialHeight", 1), 0, 1)
	compensationNode.compensationLevel = 1
	compensationNode.lastPosition = 0.5
	compensationNode.animationParts = {}

	xmlFile:iterate(key .. ".animationPart", function (index, partKey)
		local animationPart = {
			node = xmlFile:getValue(partKey .. "#node", nil, self.components, self.i3dMappings)
		}

		if animationPart.node ~= nil then
			animationPart.isLeft = xmlFile:getValue(partKey .. "#isLeft", false)
			animationPart.rotMin = xmlFile:getValue(partKey .. "#rotMin", nil, true)
			animationPart.rotMax = xmlFile:getValue(partKey .. "#rotMax", nil, true)
			animationPart.transMin = xmlFile:getValue(partKey .. "#transMin", nil, true)
			animationPart.transMax = xmlFile:getValue(partKey .. "#transMax", nil, true)

			if animationPart.rotMin ~= nil and animationPart.rotMax ~= nil or animationPart.transMin ~= nil and animationPart.transMax ~= nil then
				animationPart.lastAlpha = -1

				table.insert(compensationNode.animationParts, animationPart)
			else
				Logging.xmlWarning(xmlFile, "Failed to load slope compensation animation part '%s'. Missing values.", partKey)
			end
		else
			Logging.xmlWarning(xmlFile, "Failed to load slope compensation animation part '%s'. Missing node.", partKey)
		end
	end)

	function compensationNode.slopeDetectionCallback(_, hitObjectId, x, y, z, distance)
		if getRigidBodyType(hitObjectId) ~= RigidBodyType.STATIC then
			return true
		end

		local spec = self[SlopeCompensationExtended.SPEC_TABLE_NAME]
		spec.lastRaycastDistance = distance

		return false
	end

	return true
end

function SlopeCompensationExtended:updateSlopeCompensationNodeExt(compensationNode, position, dt)
	for _, animationPart in ipairs(compensationNode.animationParts) do
		local alpha = nil

		if animationPart.isLeft then
			alpha = 1 - MathUtil.clamp((position - 0.5) / 0.5 + compensationNode.initialHeightOffset, 0, 1)
		else
			alpha = MathUtil.clamp(position / 0.5 - compensationNode.initialHeightOffset, 0, 1)
		end

		alpha = MathUtil.lerp(compensationNode.inActiveHeight, alpha, compensationNode.compensationLevel)
		animationPart.lastAlpha = alpha

		if animationPart.rotMin ~= nil then
			local x, y, z = MathUtil.vector3ArrayLerp(animationPart.rotMin, animationPart.rotMax, alpha)

			setRotation(animationPart.node, x, y, z)

			if self.setMovingToolDirty ~= nil then
				self:setMovingToolDirty(animationPart.node)
			end
		end

		if animationPart.transMin ~= nil then
			local x, y, z = MathUtil.vector3ArrayLerp(animationPart.transMin, animationPart.transMax, alpha)

			setTranslation(animationPart.node, x, y, z)

			if self.setMovingToolDirty ~= nil then
				self:setMovingToolDirty(animationPart.node)
			end
		end
	end
end

function SlopeCompensationExtended:getCompensationGroundPositionExt(compensationNode, wheelId)
	local spec = self[SlopeCompensationExtended.SPEC_TABLE_NAME]
	local x, y, z = getWorldTranslation(compensationNode["wheel" .. wheelId .. "Node"])
	spec.lastRaycastDistance = 0

	raycastAll(x, y, z, 0, -1, 0, "slopeDetectionCallback", compensationNode.raycastDistance, compensationNode, SlopeCompensationExtended.SLOPE_COLLISION_MASK)

	local distance = spec.lastRaycastDistance

	if distance == 0 then
		distance = compensationNode["lastDistance" .. wheelId]
	else
		compensationNode["lastDistance" .. wheelId] = spec.lastRaycastDistance
	end

	return x, y - distance, z, distance ~= 0
end

function SlopeCompensationExtended:updateDebugValues(values)
	if self.isServer then
		local spec = self[SlopeCompensationExtended.SPEC_TABLE_NAME]

		for i, compensationNode in ipairs(spec.nodes) do
			table.insert(values, {
				name = string.format("node (%d)", i),
				value = string.format("%.2f", compensationNode.lastPosition)
			})

			for j, animationPart in ipairs(compensationNode.animationParts) do
				table.insert(values, {
					name = string.format("animNode (%s)", getName(animationPart.node)),
					value = string.format("%.2f", animationPart.lastAlpha)
				})
			end
		end
	end
end
